#pragma TextEncoding="UTF-8"
#pragma rtGlobals=3
#pragma version=1.40
#pragma IgorVersion=8
#pragma ModuleName=thermocouple

// --------------------- Project Updater header ----------------------
// If you have installed or updated this file using the IgorExchange
// Projects Installer (http: www.igorexchange.com/project/Updater) you can
// be notified when new versions of this project are released.
static constant kProjectID=21411 // the project node on IgorExchange
static strconstant ksShortTitle="Thermocouple" // the project short title on IgorExchange

// Thermocouple Calculator
// Tony Withers, 11/12/02
// GUI to convert thermocouple emf to temperature
// Cold junction correction is included, and wave assignments can be made
// 1.10 12/12/02 added option to make new wave
// 1.20 1/11/13  added type C, restructured code a little
// 1.30 4/9/19  cleanup, changed the complicated limit switching for setvars.
// 1.40 22/09/20 added types T, E, J, N
// comments, suggestions, improvements: https://www.wavemetrics.com/user/tony

// To add a new thermocouple type:
// 1. provide a new T2V function using one of the existing thermocouple
//    types as a template (e.g., K_T2V)
// 2. edit string constant ksTypes
// 3. edit limits function

// from the commandline:
// T2V(type, temperature)
// V2T(type, voltage)
// voltage in V, temperature in °C.
// type is a string
// print V2T("K", 1e-3) prints the temperature measured by a type K 
// thermocouple when the emf at a 0°C cold junction is 1 mV 

static strconstant ksTypes="S;B;R;K;T;E;J;N;C;D;" // list of thermocouple types for which we have functions coded

menu "Macros"
	"Thermocouple Calculator", /Q, thermocouple#ThermocoupleCalculator()
end

static function ThermocoupleCalculator()
				
	KillWindow /Z TC_Calculator
	
	// build the panel
	NewPanel /N=TC_Calculator/K=1/W=(400,200,660,375) as "Thermocouple Calculator"
	ModifyPanel /W=TC_Calculator noedit=1
	string typelist="\""+ksTypes+"\""
	PopupMenu popupTCtype,pos={30,20},size={210,21},Proc=thermocouple#TCPopMenuProc,title="Thermocouple type"
	PopupMenu popupTCtype,mode=1,popvalue="D",value=#typelist, fsize=12, bodyWidth=50
	SetVariable setvarCJT,pos={30,50},size={210,16},title="Cold junction temperature (°C)", bodyWidth=50
	SetVariable setvarCJT,value=_NUM:0, Proc=thermocouple#Recalculate, fsize=12, limits={0, 50, 1}
	SetVariable setvarEMF,pos={30,80},size={210,16},title="EMF (mV)", bodyWidth=80
	SetVariable setvarEMF,value=_NUM:0, Proc=thermocouple#Recalculate, fsize=12, limits={-50,50,0.01} // in mV
	SetVariable setvarTemp,pos={30,110},size={210,16},title="Temperature (°C)", bodyWidth=80
	SetVariable setvarTemp,value=_NUM:0, Proc=thermocouple#Recalculate, fsize=12,format="%0.1f",limits={-500, 2500, 1}
	Button buttonWaveAssignment,pos={110,140},size={130,20},Proc=thermocouple#MakeTCwavePanel,title="Wave Assignment..."
	resetLimits()
end

// function to set temperature and voltage limits to correct range for
// selected thermocouple type
static function resetLimits()
	
	ControlInfo /W=TC_Calculator setvarEMF
	variable emf=v_value
	ControlInfo /W=TC_Calculator setvarCJT
	variable CJT=v_value
	ControlInfo /W=TC_Calculator popupTCtype
	string tp=S_Value
	wave w=limits(s_value) // get two point wave containing T limits for t/c type
	SetVariable setvarTemp win=TC_Calculator, limits={w[0],w[1],1}
	SetVariable setvarEMF win=TC_Calculator, limits={(T2V(tp, w[0])-T2V(tp, CJT))*1e3,(T2V(tp, w[1])-T2V(tp, CJT))*1e3,0.01}
end

static function TCPopMenuProc(STRUCT WMPopupAction &s)
	resetLimits()
	recalculateT()
	return 0
end

static function Recalculate(STRUCT WMSetVariableAction &s)
	if ( stringmatch (s.ctrlName, "*Temp") ) // T->emf calculation
		recalculateEMF()
	else // emf or CJT changed; emf->T calculation
		resetLimits()
		recalculateT()
	endif
	return 0
end

static function recalculateT()
	
	ControlInfo /W=TC_Calculator setvarEMF
	variable emf=v_value
	ControlInfo /W=TC_Calculator setvarCJT
	variable CJT=v_value
	ControlInfo /W=TC_Calculator popupTCtype
	string TC=S_Value
	SetVariable  setvarTemp win=TC_Calculator, value=_NUM:V2T(TC, emf*1e-3 + T2V(TC, CJT))
end

static function recalculateEMF()

	ControlInfo /W=TC_Calculator setvarCJT
	variable CJT=v_value
	ControlInfo /W=TC_Calculator setvarTemp
	variable temperature=v_value
	ControlInfo /W=TC_Calculator popupTCtype
	string TC=S_Value
	SetVariable  setvarEMF win=TC_Calculator,value=_NUM:1e3*(T2V(TC, temperature) - T2V(TC, CJT)) // V -> mV
end

static function MakeTCwavePanel(STRUCT WMButtonAction &s)

	if (s.eventCode!=2)
		return 0
	endif
	
	DoWindow /F TCWavePanel
	if (V_flag)
		return 0
	endif

	GetWindow /Z TC_Calculator wsize
	NewPanel /K=1/W=(V_left,V_bottom,V_left+390,V_bottom+170)/N=TCWavePanel as "Thermocouple wave assignment"
	ControlInfo /W=TC_Calculator popupTCtype
	PopupMenu popupType,pos={10,20},size={150,21},title="Thermocouple type", Proc=thermocouple#TCWavePopMenuProc
	PopupMenu popupType,mode=1,popvalue=S_Value,value=ksTypes
	PopupMenu popupUnit,pos={10,50},size={156,21},title="Source wave unit"
	PopupMenu popupUnit,mode=1,popvalue="volts",value="volts;mV;microvolts;"
	PopupMenu popupUnit,Proc=thermocouple#TCWavePopMenuProc
	ControlInfo /W=TC_Calculator setvarCJT
	SetVariable setvarCJT,pos={10,80},size={120,16},title="Cold junction (°C)"
	SetVariable setvarCJT,limits={0,50,1},value=_NUM:v_value, Proc=thermocouple#TCSetVarProc
	
	PopupMenu popupSource,pos={200,20},size={132,21},title="Source wave"
	PopupMenu popupSource,mode=1,value=WaveList("*",";", "DIMS:1")
	PopupMenu popupSource, Proc=thermocouple#TCWavePopMenuProc
	PopupMenu popupDestination,pos={200,50},size={151,21},title="Destination wave"
	PopupMenu popupDestination,value= "new...;"+WaveList("*",";", "DIMS:1"),mode=1
	PopupMenu popupDestination,Proc=thermocouple#TCWavePopMenuProc
	SetVariable setvarNewName,pos={200,80},size={180,16},value=_STR:UniqueName("Wave", 1, 0)
	SetVariable setvarNewName, title="New wave name:" , Proc=thermocouple#TCSetVarProc
	SetVariable setvarCmd pos={10,110}, size={370,60}, value=_STR:"", noedit=1, valueBackColor=(60067,60075,60063)
	Button buttonClip,pos={75,140},size={50,20},title="To clip", Proc=thermocouple#DoTCWaveAssignment
	Button buttonGo,pos={10,140},size={50,20},title="Do it", Proc=thermocouple#DoTCWaveAssignment
	
	BuildWaveAssignmentCmd() // set the cmd string
	SetWindow TCWavePanel, hook=thermocouple#HookTCWavePanel
end

// window hook prevents wave assignment command from becoming stale
static function HookTCWavePanel(STRUCT WMWinHookStruct &s)
	if(s.eventCode==0)
		ControlUpdate /A /W=TCWavePanel
		BuildWaveAssignmentCmd()
	endif
	return 0
end

static function TCWavePopMenuProc(STRUCT WMPopupAction &s)
	if(s.eventCode!=-1)
		BuildWaveAssignmentCmd()
	endif
	return 0
end

static function TCSetVarProc(STRUCT WMSetVariableAction &s)
	if(s.eventCode!=-1)
		BuildWaveAssignmentCmd()
	endif
	return 0
end

static function BuildWaveAssignmentCmd()
	
	string cmd
	variable OK=1
	ControlInfo /W=TCWavePanel setvarNewName
	string NewWaveName=s_value
	ControlInfo /W=TCWavePanel setvarCJT
	variable CJT=v_value
	ControlInfo /W=TCWavePanel  popupDestination
	if (V_value==1) // make new wave
		SetVariable setvarNewName, win=TCWavePanel, disable=0
		ControlInfo /W=TCWavePanel  popupSource
		sprintf cmd, "duplicate %s, %s; %s =", S_Value, CleanupName(NewWaveName, 0), CleanupName(NewWaveName, 0)
		if (CheckName(CleanupName(NewWaveName, 0), 1))
			OK=0
		endif
	else
		SetVariable setvarNewName, win=TCWavePanel, disable=1
		sprintf cmd, "%s =", S_Value
		if (exists(S_Value))
			wave dest=$S_Value
		else
			OK=0
		endif
	endif
	
	ControlInfo /W=TCWavePanel  popupType
	sprintf cmd, "%s V2T(\"%s\", " cmd, S_Value
	ControlInfo /W=TCWavePanel  popupSource
	sprintf cmd, "%s%s", cmd, S_Value
	if (exists(S_Value))
		wave source=$S_Value
	else
		OK=0
	endif
	ControlInfo /W=TCWavePanel popupUnit
	if (V_Value-1)
		sprintf cmd, "%s*1e-%d", cmd, (V_Value-1)*3
	endif
	if (CJT)
		ControlInfo /W=TCWavePanel  popupType
		sprintf cmd, "%s + T2V(\"%s\", %g)" cmd, S_Value, CJT
	endif
	cmd+=")"
	SetVariable setvarCmd win=TCWavePanel, value=_STR:cmd
	Button buttonGo, win=TCWavePanel, disable=2*(OK==0)
end

static function DoTCWaveAssignment(STRUCT WMButtonAction &s)

	if (s.eventCode!=2)
		return 0
	endif
		
	ControlInfo /W=TCWavePanel setvarCmd
	string cmd=s_value
	
	if (stringmatch(s.ctrlName, "buttonClip"))
		cmd=ReplaceString("\r", cmd, "; ") // replace carriage return with ;
		// keep this in case we switch to embedded notebook
		PutScrapText cmd
		return 0
	endif
	
	// do it button
	ControlInfo /W=TCWavePanel setvarNewName
	string NewWaveName=s_value
	ControlInfo /W=TCWavePanel  popupDestination
	if (V_value>1) // destination is a pre-existing wave
		wave dest=$S_Value
		string AlertStr=S_Value
		ControlInfo /W=TCWavePanel popupSource
		wave source=$S_Value
		if (stringmatch(AlertStr, S_Value))
			DoAlert 1, "Same source and destination. Overwrite "+AlertStr+"?"
			if (V_Flag>1)
				return 0
			endif
		endif
		if (DimSize(source, 0)!=DimSize(dest, 0)) // different length waves
			AlertStr+= " and "+S_Value+" are not the same length. Continue?"
			DoAlert 1, AlertStr
			if (V_Flag>1)
				return 0
			endif
		endif
	endif
	string cmdItem
	variable i
	for (i=0; i<ItemsInList(cmd, "\r"); i+=1) // execute each line of cmd
		cmdItem=StringFromList(i, cmd  , "\r")
		Execute cmdItem
		Print cmdItem // commentise this line to prevent writing commands to history
	endfor

	// need to update new wave name suggestion if we've used the original one
	SetVariable setvarNewName,win=TCWavePanel,value=_STR:UniqueName("Wave", 1, 0)
	BuildWaveAssignmentCmd()
end

// provides the working temperature range in degrees C 
static function /wave limits(string type)	
	Make /free/n=2 w
	strswitch(type)
		case "S":
			w={-50,1768}
			break
		case "B":
			w={0,1820}
			break
		case "R":
			w={-50,1768}
			break
		case "K":
			w={-270,1372}
			break
		case "T":
			w={-270,400}
			break
		case "E":
			w={-270,1000}
			break
		case "J":
			w={-210,1200}
			break
		case "N":
			w={-270,1300}
			break
		case "C":
			w={0,2320}
			break
		case "D":
			w={0,2320}
			break
	endswitch
	return w
end

// voltage in V, returns temperature in deg C
function V2T(string type, variable voltage)
	Make /free/n=(strlen(type)) w = char2num(type[p])
	wave lim=limits(type)
	FindRoots /L=(lim[0]-1)/H=(lim[1]+1)/Q/T=0.0001/Z=(voltage) thermocouple#T2V_wrapper, w
	return v_Flag ? NaN : round(10000*V_root)/10000
end

// temperature in deg C, returns voltage in V
function T2V(string type, variable temperature)
	FUNCREF T2Vtemplate T2V = $"thermocouple#"+type+"_T2V"
	return T2V(temperature)
end

// function with correct format for findroots
// w contains ASCII representing thermocouple type
static function T2V_wrapper(wave w, variable t90)
	string type=""
	variable i
	for(i=0;i<numpnts(w);i++)
		type+=num2char(w[i])
	endfor
	FUNCREF T2Vtemplate T2V = $"thermocouple#"+type+"_T2V"
	return T2V(t90)
end

function T2Vtemplate(variable t)
end

// ITS-90 thermocouple functions from NIST Monograph # 175:
// Temperature-Electromotive force reference functions and tables for
// the letter-designated thermocouple types based on the ITS-90
// G.W. Burns, M.G. Scroger and G.F. Strouse (1993)

// Types C and D are from ASTM standard E 988-96:
// Standard Temperature-Electromotive Force (EMF) Tables
// for Tungsten-Rhenium Thermocouples

// ----------------------  type K thermocouple   ----------------------------

static function K_T2V(variable t90) // returns voltage (volts) as functon of T (degC)
	Make /D/free/n=1 c
	variable EMF=0
	if (t90==0)
		return 0
	endif
	if (t90<0)
		c={0,3.9450128025e-2,2.3622373598e-5,-3.2858906784e-7,-4.9904828777e-9,-6.7509059173e-11}
		c[6]={-5.7410327428e-13,-3.1088872894e-15,-1.0451609365e-17,-1.9889266878e-20,-1.6322697486e-23}
	else //t90>0
		c={-1.7600413686e-2,3.8921204975e-2,1.8558770032e-5,-9.9457592874e-8,3.1840945719e-10}
		c[5]={-5.6072844889e-13,5.6075059059e-16,-3.2020720003e-19,9.7151147152e-23,-1.2104721275e-26,0}
		EMF=1.185976e-1*exp(-1.183432e-4*(t90-126.9686)^2)
	endif
	EMF += poly(c, t90)
	return EMF*1e-3 // mV -> volts
end

// ----------------------  type S thermocouple   ----------------------------

static function S_T2V(variable t90) // returns voltage (volts) as functon of T (degC)
	Make /D/free/N=1 c
	if (t90==0)
		return 0
	elseif (t90<1064.18)
		c={0,5.40313308631e-3,1.25934289740e-5,-2.32477986889e-8,3.22028823036e-11}
		c[5]={-3.31465196389e-14,2.55744251786e-17,-1.25068871393e-20,2.71443176145e-24}
	elseif(t90>1664.5)
		c={1.46628232636e2,-2.58430516752e-1,1.63693574641e-4,-3.30439046987e-8,-9.43223690612e-15}
	else // 1664.5>=t90>=1064.18
		c={1.32900444085,3.34509311344e-3,6.54805192818e-6,-1.64856259209e-9,1.29989605174e-14}
	endif
	return poly(c,t90)*1e-3 // mV -> volts
end

// ----------------------  type B thermocouple   ----------------------------

static function B_T2V(variable t90) // returns voltage (volts) as functon of T (degC)
	Make /D/free/N=1 c
	if (t90==0)
		return 0
	elseif (t90<630.615)
		c={0,-2.4650818346e-4,5.9040421171e-6,-1.3257931636e-9,1.5668291901e-12,-1.6944529240e-15,6.2990347094e-19}
	else //  t90>=630.615
		c={-3.8938168621,2.8571747470e-2,-8.4885104785e-5,1.5785280164e-7,-1.6835344864e-10}
		c[5]={1.1109794013e-13,-4.4515431033e-17,9.8975640821e-21,-9.3791330289e-25}
	endif
	return poly(c,t90)*1e-3 // mV -> volts
end

// ----------------------  type C thermocouple   ----------------------------

// from ASTM standard E 988-96 (2002)
static function C_T2V(variable temperature) // returns voltage (volts) as functon of T (degC)
	Make /D/free/n=1 c
	if(temperature<630.615)
		c={0,1.3406032e-2,1.1924992e-5,-7.9806354e-9,-5.0787515e-12,1.3164197e-14,-7.9197332e-18}
	else
		c={4.0528823e-1,1.1509355e-2,1.5696453e-5,-1.3704412e-8,5.2290873e-12,-9.2082758e-16,4.5245112e-20}
	endif
	return poly(c,temperature)*1e-3 // mV -> volts
end

// ----------------------  type D thermocouple   ----------------------------

// from ASTM standard E 988-96 (2002)
static function D_T2V(variable temperature) // returns voltage (volts) as functon of T (degC)
	Make /D/free/N=1 c
	if (temperature<783)
		 c={0,9.5921929e-3,2.0068371e-5,-1.3786121e-8,-1.1620542e-11,3.9875300e-14,-4.2429757e-17,1.6821225e-20}
	else
		 c={2.2097354,-1.4500612e-3,4.2898234e-5,-4.2816409e-8,2.4132609e-11,-8.1885541e-15,1.5873209e-18,-1.4320975e-22}
	endif
	return poly(c,temperature)*1e-3 // mV -> volts
end

// ----------------------  type R thermocouple   ----------------------------

static function R_T2V(variable t90) // returns voltage (volts) as functon of T (degC)
	Make /D/free/N=1 c
	if (t90==0)
		return 0
	elseif (t90<1064.18)
		c={0,5.28961729765e-3,1.39165589782e-5,-2.38855693017e-8,3.56916001063e-11,-4.62347666298e-14}
		c[6]={5.00777441034e-17,-3.73105886191e-20,1.57716482367e-23,-2.81038625251e-27}
	elseif (t90>1664.5)
		c={1.52232118209e2,-2.68819888545e-1,1.71280280471e-4,-3.45895706453e-8,-9.34633971046e-15}
	else //  1664.5>=t90>=1064.18
		c={2.95157925316,-2.52061251332e-3,1.59564501865e-5,-7.64085947576e-9,2.05305291024e-12,-2.93359668173e-16}
	endif
	return poly(c,t90)*1e-3 // mV -> volts
end

// ----------------------  type T thermocouple   ----------------------------

static function T_T2V(variable t90) // returns voltage (volts) as functon of T (degC)
	Make /D/free/N=1 c
	if (t90==0)
		return 0
	elseif (t90<0)
		c={0,3.8748106364e-2,4.4194434347e-5,1.1844323105e-7,2.0032973554e-8}
		c[5]={9.0138019559e-10,2.2651156593e-11,3.6071154205e-13,3.8493939883e-15,2.8213521925e-17}
		c[10]={1.4251594779e-19,4.8768662286e-22,1.0795539270e-24,1.3945027062e-27,7.9795153927e-31}
	else //  400>=t90>0
		c={0,3.8748106364e-2,3.3292227880e-5,2.0618243404e-7,-2.1882256846e-9}
		c[5]={1.0996880928e-11,-3.0815758772e-14,4.5479135290e-17,-2.7512901673e-20}
	endif
	return poly(c,t90)*1e-3 // mV -> volts
end

// ----------------------  type E thermocouple   ----------------------------

static function E_T2V(variable t90) // returns voltage (volts) as functon of T (degC)
	Make /D/free/N=1 c
	if (t90==0)
		return 0
	elseif (t90<0)
		c={0,0.58665508708e-1,0.45410977124e-4,-0.77998048686e-6,-0.25800160843e-7}
		c[5]={-0.59452583057e-9,-0.93214058667e-11,-0.10287605534e-12,-0.80370123621e-15,-0.43979497391e-17}
		c[10]={-0.16414776355e-19,-0.39673619516e-22,-0.55827328721e-25,-0.34657842013e-28}
	else //  1000>=t90>0
		c={0,0.58665508710e-1,0.45032275582e-4,0.28908407212e-7,-0.33056896652e-9}
		c[5]={0.65024403270e-12,-0.19197495504e-15,-0.12536600497e-17,0.21489217569e-20,-0.14388041782e-23}
		c[10]={0.35960899481e-27}
	endif
	return poly(c,t90)*1e-3 // mV -> volts
end

// ----------------------  type J thermocouple   ----------------------------

static function J_T2V(variable t90) // returns voltage (volts) as functon of T (degC)
	Make /D/free/N=1 c
	if (t90==0)
		return 0
	elseif (t90<760)
		c={0, 0.50381187815e-1, 0.30475836930e-4,-0.85681065720e-7, 0.13228195295e-9}
		c[5]={-0.17052958337e-12, 0.20948090697e-15,-0.12538395336e-18, 0.15631725697e-22}
	else //  1200>=t90>760
		c={0.29645625681e3,-0.14976127786e1, 0.31787103924e-2,-0.31847686701e-5,0.15720819004e-8}
		c[5]={-0.30691369056e-12}
	endif
	return poly(c,t90)*1e-3 // mV -> volts
end

// ----------------------  type N thermocouple   ----------------------------

static function N_T2V(variable t90) // returns voltage (volts) as functon of T (degC)
	Make /D/free/N=1 c
	if (t90==0)
		return 0
	elseif (t90<0)
		c={0, 0.26159105962e-1, 0.10957484228e-4,-0.93841111554e-7,-0.46412039759e-10}
		c[5]={-0.26303357716e-11,-0.22653438003e-13,-0.76089300791e-16,-0.93419667835e-19}
	else //  1300>=t90>0
		c={0, 0.25929394601e-1, 0.15710141880e-4, 0.43825627237e-7,-0.25261169794e-9}
		c[5]={0.64311819339e-12,-0.10063471519e-14, 0.99745338992e-18,-0.60863245607e-21, 0.20849229339e-24}
		c[10]={-0.30682196151e-28}
	endif
	return poly(c,t90)*1e-3 // mV -> volts
end